/*
	Copyright (C) 2011 Micha Hersch, University of Lausanne, Swiss Institute of Bioinformatics

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*!
 *	\file snakeNest.cpp
 *	\author Micha Hersch
 *	\date
 */

#include <fstream>
#include "snakeNest.h"

using namespace std;

#define WITH_BASE


#ifdef TRAINING
//using namespace cv;
  
#define AXIS_LEN 10
#endif

// length of executable file name
#ifdef _WIN32
#define EXEC_LENGTH 12 
#else
#define EXEC_LENGTH 8
#endif

#ifdef WITH_BASE
#define EYECUP -1
#else
#define EYECUP 1
#endif

SnakeNest::SnakeNest(int hor, int ver,int length){
  nrow = ver;
  ncol = hor;
  nb = ncol*nrow;
  nb_points = length;
  snakes = new Snake[nb];
  grounds = new int[ver];
  for(int i=0;i<ver;i++){grounds[i]=0;}
  grabber=NULL;
  img=NULL;
  disp=NULL;
  disp_bak=NULL;
  idisp=NULL;
  wname[0] = 0;
  selected_point = -1;
  selected_snake = -1;
  more_selected_points = -1;
  bw_thresholds = new int[nb];
  for(int i=0;i<nb;i++){bw_thresholds[i]=-1;}
  global_threshold = -1;
  moving_snake = -1;
  saveImages=0;
  keep_trace=0;
  oriwin = 0;
  win_factor=-1;
  zoom_factor=1;
  thresh_w=200;
  thresh_h=200;
  strcpy(imgfileformat,".tif");
  strcpy(baseimgfile,"results/track");
  state =NONE;
  zooming= 0;
  SnakeNest::instance = this;
}
SnakeNest::~SnakeNest(){
   IMKILL(img);
  IMKILL(old_img);
  IMKILL(disp);
  IMKILL(idisp);
  IMKILL(disp_bak);
#ifdef GLOBAL_PREPROCESSING
  IMKILL(imgeq);
  IMKILL(old_imgeq);
#endif
  closeDisplay();
  //  cout<<"deleting snakes"<<endl;
  delete[] snakes;
  // cout<<"deleting grounds"<<endl;
    delete[] grounds;
    delete[] bw_thresholds; // can cause some invalid pointer error (why?)
    // cout<<"done"<<endl;
}


void SnakeNest::initGrabber(const char *fname){
    if(grabber) grabber->Kill();
    grabber = new ImgGrabber(fname);
    grabber->SetLoop(0);
    img = cvCreateImage(grabber->GetSize(),8,1);
    zoomRect = cvRect(0,0,img->width,img->height);
    // cout<<"width "<<img->width<<endl;
    // old_img = cvCreateImage(grabber->GetSize(),8,1);
    old_img = cvCloneImage(img);
    grabber->GrabFrame(&disp);
    //grabber->SetFrame(1);
    grabber->Skip(-1);
#ifdef GLOBAL_PREPROCESSING
    imgeq = cvCreateImage(grabber->GetSize(),8,1);
    old_imgeq = cvCloneImage(imgeq);
#endif
}


void SnakeNest::basicInitSnakes(int quincunx){
  int hsize,vsize;
  vsize = img->height/nrow;
  if(quincunx){
      hsize= (int)(img->width/(ncol+0.5));
  }
  else{
    hsize= img->width/ncol;
  }
  for (int i=0;i<ncol;i++){
    for(int j=0;j<nrow;j++){
      int offset = (quincunx && !(j%2))?0.5*hsize:0;
      int cx = i*hsize+offset+hsize/2 ;
      int cy = j*vsize+vsize/2  ;
      getSnake(i,j)->drop(cx,cy,nb_points,30);
    }
    cout<<endl;
  }
}


void SnakeNest::initSnakes(int quincunx){
#ifdef WITH_EQUALIZATION_INIT
  IplImage *imeq= cvCreateImage(cvSize(img->width,img->height),8,1);
  cvEqualizeHist(img,imeq);
#else 
    IplImage *imeq =img;
#endif
    // tiling 
    int hsize,vsize;
    vsize = img->height/nrow;
    if(quincunx){
      hsize= (int)(img->width/(ncol+0.5));
    }
    else{
      hsize= img->width/ncol;
    }
    
     for (int i=0;i<ncol;i++){
     for(int j=0;j<nrow;j++){
    //    for (int i=0;i<1;i++){
    //   for(int j=0;j<1;j++){

      int offset = (quincunx && !(j%2))?0.5*hsize:0;
      CvRect frame = cvRect(i*hsize+offset,j*vsize,hsize,vsize);
      //  std::cout<<i<<" "<<j<<std::endl;
      getSnake(i,j)->init(imeq,nb_points,&frame,detect/(float)max_detect,grounds[j]);
    }
  }
#ifdef WITH_EQUALIZATION_INIT
  IMKILL(imeq);
#endif
}


void SnakeNest::initDisplay(const char *win_name,int autosize,int trackbar){
  disp_bak = cvCloneImage(disp); 
  idisp = cvCloneImage(disp);
  strcpy(wname,win_name);
  cout<<"image size "<<disp->width<<" "<<disp->height<<endl;
   cvNamedWindow(wname,autosize);
  cvMoveWindow(wname,100,100);
 //  cvShowImage(wname,disp);waitForCommand(0,0);
 
  cvSetMouseCallback(wname,on_mouse,this);
  switch(trackbar){
  case 1:
    cvCreateTrackbar("detection",win_name,&detect,max_detect,detect_tbar);
    break;
  case 2:
    cvCreateTrackbar("smoothness",win_name,&modulation,max_modulation,modulation_tbar);
    break;
  default:break;
  }
  // if(autosize==CV_WINDOW_NORMAL){
    cvMoveWindow(wname,100,100);
    if(win_factor <0){
      if(disp->width>500 || disp->height>500){
	win_factor = min(500/(float)disp->width,500/(float)disp->height);
      }
    }
    //    cout<<"zoom factor "<<win_factor<<endl;
    cvResizeWindow(wname,win_factor*disp->width,win_factor*disp->height);

    // }

//   cvNamedWindow("debug1",0);
//   cvNamedWindow("debug2",0);
}



void SnakeNest::closeDisplay(){
  cvCopy(disp_bak,disp);
  cvDestroyWindow(wname);
  IMKILL(disp_bak);
  IMKILL(idisp);
}


int SnakeNest::nextFrame(){
  if(grabber->IsDone()){return 0;}
  //  if(disp_bak)cvCvtColor(disp_bak,old_img,CV_BGR2GRAY);  
  if(zooming){unzoom(0);}
   grabber->GrabFrame(&disp);
  // if(zoomRect.width<disp->width || zoomRect.height<disp->height) {
  //   cvSetImageROI(disp,zoomRect);
  // }
#ifdef GLOBAL_PREPROCESSING
   if(disp_bak)cvCvtColor(disp_bak,old_imgeq,CV_BGR2GRAY);
  cvCvtColor(disp,imgeq,CV_BGR2GRAY);

#ifdef WITH_EQUALIZATION
  if(disp_bak)cvEqualizeHist(old_imgeq,old_img);
  cvEqualizeHist(imgeq,old_img);
#endif

#ifdef WITH_GLOBAL_THRESHOLDING
 
  if(disp_bak)  cvThreshold(old_imgeq,old_img,bw_threshold,255,CV_THRESH_BINARY);
  // cout<<"thresh2"<<imgeq->size.width<<" " <<imgeq->size.height<<endl;
  // cout<<"thresh2"<<img->size.width<<" " <<img->size.height<<endl;
  cvThreshold(imgeq,img,bw_threshold,255,CV_THRESH_BINARY);
#endif

#ifdef WITH_LOCAL_THRESHOLDING
   if(disp_bak) cvCopy(old_imgeq,old_img);
   cvCopy(imgeq,img);
 #endif

#else //NO PREPROCESSING
  if(disp_bak)cvCvtColor(disp_bak,old_img,CV_BGR2GRAY);  
  cvCvtColor(disp,img,CV_BGR2GRAY);
#endif
  cvCopy(disp,disp_bak);
  return 1;
}


int SnakeNest::closestSnake(Vec2 v,int bot,int exclude){
  float mindist=(img->height+img->width)*(img->height+img->width);
  int mini = 0;
  int only_valid = (exclude==-2);
  for(int i=0;i<nb;i++){
    if(i!=exclude){
      Snake *s = getSnake(i);
      if(!only_valid || s->isValid()){ 
	int si= s->getSize();
	if(si){
	  Vec2 p=s->getPoint(bot?si-1:0);
	  p -=v; 
	  float dist = p.lengthSquared();
	  if(dist<mindist){
	    mindist = dist;
	    mini = i;
	  }
	}
      }
    }
  }
  return mini;
}

int SnakeNest::closestSleepingSnake(Vec2 v){
  float mindist=(img->height+img->width)*(img->height+img->width);
  int mini = -1;
 
  for(int i=0;i<nb;i++){
      Snake *s = getSnake(i);
      if(!s->isValid()){
        Vec2 p=s->getPoint(0);
        p -=v; 
        float dist = p.lengthSquared();
        if(dist<mindist){
          mindist = dist;
          mini = i;
        }
      }
  }
  return mini;
}



void SnakeNest::selectPoint(int x,int y){
  int s[4];
  int p[4];
  int dist[4];
  int nb = 4;
  int mindist=(img->width+img->height)*(img->width+img->height);
  int mini = 0;
  Vec2 v(x,y);

  // finding the (at most) four closest snake extremities
  s[0]=closestSnake(v,0);
  s[1]=closestSnake(v,0,s[0]);
  s[2]=closestSnake(v,1);
  s[3]=closestSnake(v,1,s[2]);
 
  //removing double occurences in s
  for(int i=0;i<nb;i++){
    for(int j=i+1;j<nb;j++){
      if(s[i]==s[j]){
        s[j] = s[--nb];
      }
    }
  }
  // find the closest point in each snake
  for(int i=0;i<nb;i++){
    p[i] = getSnake(s[i])->findClosestPoint(Vec2(x,y),dist+i);
    if(mindist>dist[i]){
      mindist=dist[i];
      mini = i;
    }
  }
  selected_snake=s[mini];
  selected_point=p[mini];
}


void SnakeNest::selectManyPoints(int x,int y){
  selectPoint(x,y);
  Snake *s = getSnake(selected_snake);
  Vec2 selpos = s->getPoint(selected_point);
  int i= selected_point-1;
  while((s->getPoint(i)- selpos).lengthSquared()<=5){
    if(!i--)break;
  }
  more_selected_points = i+1;
  //  cout<<more_selected_points<<" "<<selected_point<<endl;
}

void SnakeNest::setGround(int ground){
 
  float mindist =img->height;
  int minrow = 0;
  //looking for the closest row  
  for(int i=0;i<nrow;i++){
    int cnt = 0;
    float dist = 0;
    for(int j=0;j<ncol;j++){
      Snake *s = getSnake(j,i);
        int si = s->getSize();
        if (si){
          dist+=s->getPoint(si-1).y;
        cnt++;
        }
      }
      dist = fabs(dist/cnt-ground);
      if(dist<mindist){
        mindist=dist;
        minrow=i;
      }
  }
 grounds[minrow]=ground;
  // anchoring to ground
  for(int j=0;j<ncol;j++){
    getSnake(j,minrow)->anchorToGround(ground);
   }
  display(CV_RGB(255,0,0));
}

void SnakeNest::waitForCommand(int x, int y){
  char c=0;
  cout<<"command: ";
  cout.flush();
  c= cvWaitKey(5000);
  cout<<c<<endl;
  switch(c){
  case 0:   cout<<"no command"<<endl;break;
  case 'i':
    win_factor /= 0.8;
    cvResizeWindow(wname,win_factor*disp->width,win_factor*disp->height);
    cvShowImage(wname,disp);
    break;
  case 'o':
    cout<<wname<<endl;
    win_factor *= 0.8;
    cvResizeWindow(wname,win_factor*disp->width,win_factor*disp->height);
    cvShowImage(wname,disp);
    break;
    break;
  default:  cout<<"not a known command"<<endl;
  }
}


void SnakeNest::correct(int event,int flags,int x,int y){
 
  int c;
  Snake *s;
  
  if(event ==CV_EVENT_MOUSEMOVE){
    if(moving_snake==-1 && selected_point==-1 )return;// no need to use cpu
    if(moving_snake!=-1 && moving_snake != selected_snake)return;// no need to use cpu
  }  
  x += zoomRect.x;
  y += zoomRect.y;
  Vec2 posi(x,y);  
 if(event == CV_EVENT_LBUTTONUP){
      resetSelectedPoint(x,y);
      return;
  }
 //  cout<<"event "<<event<<" flag "<<flags<< "pos "<<x<<" "<<y<<endl;
  if(flags & CV_EVENT_FLAG_SHIFTKEY){    //SHIFT
      switch(event){
      case CV_EVENT_LBUTTONDOWN: 
	if(flags & CV_EVENT_FLAG_CTRLKEY) { // moves the whole snake
	  selectMovingSnake(x,y);}       
	else{
	  s=getSnake(closestSnake(posi,0,-2));
	  s->getCotyledon().setPosition(posi); // updates tip
	  s->setPoint(0,s->getCotyledon().getPosition());
	}
	break;
      case CV_EVENT_MOUSEMOVE:
	moveSelectedPoint(x,y);break;
/*       case CV_EVENT_LBUTTONUP:  // done moving snake */
/* 	resetSelectedPoint(x,y);break; */
      case CV_EVENT_RBUTTONUP: // reset thresholding
#ifdef WITH_BASE
	if(flags & CV_EVENT_FLAG_CTRLKEY) {
	  s=getSnake(closestSnake(posi,0,-2));
	  s->getBase().setPosition(posi); // updates base
	  s->setLastPoint(s->getBase().getPosition());
  
	}
	else
#endif	
	  {c = closestSnake(posi,0,-2);getThresholdValues(c);}
	break;
      }
    }
  else{// NOT SHIFT
    if(flags & CV_EVENT_FLAG_ALTKEY){ //ALT
	switch(event){
	case CV_EVENT_RBUTTONUP: // new snake
	  c = closestSleepingSnake(posi);
	  cout<<"new hypocotyl: "<<c<<endl;
	  if(c>=0){
	    s=getSnake(c);
	    s->validate();
	    s->setPoint(0,posi);
	    s->getCotyledon().setPosition(posi);
	    threshold(c);
	    s->downFromTip(img,20,60);
#ifdef WITH_BASE
	    s->getBase().setPosition(s->getLastPoint());
#endif
	  }
	  else{
	    cout<<"no more snakes available"<<endl;
	  }
	  break;
	case CV_EVENT_LBUTTONDOWN: //command
	  waitForCommand(x,y);
	  break;
	}
      }
      else{
	if (flags & CV_EVENT_FLAG_CTRLKEY){    
	  switch(event){
	  case CV_EVENT_LBUTTONDOWN: // moves single point
	    selectPoint(x,y);initPointMove(); break;
	  case CV_EVENT_MOUSEMOVE:
	    moveSelectedPoint(x,y);break;
/* 	case CV_EVENT_LBUTTONUP:  // done moving point */
/* 	  resetSelectedPoint(x,y);break; */
	  case CV_EVENT_RBUTTONUP: // delete snake
	    c = closestSnake(posi,0,-2);  
	    getSnake(c)->invalidate();
	    if(c==selected_snake){selected_snake=-1;}
	    break;
	  }
	}	  
	else{
	  switch(event){
	  case CV_EVENT_LBUTTONDOWN:
	    selectSnake(x,y);break;
	  case CV_EVENT_RBUTTONUP: //zooming in 
	    zoom(Vec2(x,y),0.8);
	    break;	
	  }
	}
      }
    }
  display(CV_RGB(255,0,0));
}



void SnakeNest::selectSnake(int x,int y){
  int c=closestSnake(Vec2(x,y));
  selectSnake(c);
}
void SnakeNest::selectSnake(int c){
  if(c==selected_snake){//deselecting
    selected_snake =-1;
    getSnake(c)->draw(disp,CV_RGB(255,0,0));
  }
  else{//selecting
    if(selected_snake!=-1){
      getSnake(selected_snake)->draw(disp,CV_RGB(255,0,0));
    }
    selected_snake =c;
    getSnake(c)->validate();
    getSnake(c)->draw(disp,CV_RGB(255,0,0)*0.5);
    modulation = cvRound((getSnake(c)->getEnergyBalance()*max_modulation)/max_balance);
    cvSetTrackbarPos("smoothness",wname,modulation);
  }
  cvShowImage(wname,disp);
}

void SnakeNest::tuneSelectedSnake(float mod){
   int ec = 0;
#ifdef SNAKE_EYECUP
    ec =EYECUP;
#endif
    Snake *s = selectedSnake();
    if(s){
      s->setEnergyBalance(mod);s->updateSnake(img,10,ec);
    }
}


void SnakeNest::writeCotyledons(const char *fname){
  ofstream out(fname);
  if(out.is_open()){
    for(int i=0;i<nb;i++){
      out<<i<<" "<<getSnake(i)->getCotyledon()<<endl;
    }
    out.close();
  }
  else{
    cerr<<"cannot open file"<<fname<<endl;
  }
}

void SnakeNest::readCotyledons(const char *fname){
  ifstream in(fname);
  if(in.is_open()){
      for(int i=0;i<nb;i++){
	in>>i>>getSnake(i)->getCotyledon();
      }
      in.close();
  }
  else{
    cerr<<"cannot open file"<<fname<<endl;
  }
}

void SnakeNest::display(CvScalar col){
  try{
    cvCopy(disp_bak,disp);

  }
  catch(cv::Exception e){
    cout<<"Exception: "<<e.code<<" "<<e.code<<endl;
    cout<<"disp_bak: "<<disp_bak->depth<<" "<<disp_bak->width<<" "<<disp->height<<endl;
    cout<<"disp: "<<disp->depth<<" "<<disp->width<<" "<<disp->height<<endl;
  }
  for(int i=0;i<nb;i++){
    if(selected_snake==i){
      getSnake(i)->draw(disp,0.6*col);
      getSnake(i)-> drawCotyledon(disp);
#ifdef WITH_BASE
      getSnake(i)-> drawBase(disp);
#endif
    }
    else{
      if(!getSnake(i)->isValid()){
	//	getSnake(i)->draw(disp,0.3*col);
      }
      else{
	getSnake(i)->draw(disp,col);
      }
      if(getSnake(i)->isValid()){
	getSnake(i)-> drawCotyledon(disp);
#ifdef WITH_BASE
	getSnake(i)-> drawBase(disp);
#endif
      }    
    }
  }
  if(zooming){
    cvSetImageROI(disp,zoomRect);
    cvSetImageROI(disp_bak,zoomRect);
  }
  if(wname[0])cvShowImage(wname,disp);
}


void SnakeNest::showOrientationLength(){
  for(int i=0;i<nb;i++){
    if(getSnake(i)->isValid()){
      getSnake(i)->drawLength(disp,orient_length,CV_RGB(0,0,255));
    } 
  }
  cvShowImage(wname,disp);
}

void SnakeNest::bwUpdateSnake(int i, int nbit,int eyecup){
  threshold(i);
  getSnake(i)->updateCotyledon(img,old_img);
  Vec2 pos = getSnake(i)->getCotyledon().getPosition();
  if(pos.x>=0 &&pos.y >=0){
    getSnake(i)->setPoint(0,getSnake(i)->getCotyledon().getPosition());
  }
  else{
    pos= getSnake(i)->getPoint(0);
    getSnake(i)->getCotyledon().setPosition(pos);
  }
#ifdef WITH_BASE
  //  getSnake(i)->updateBase(img,old_img);
  pos = getSnake(i)->getBase().getPosition();

  if(pos.x>=0 &&pos.y >=0){
    getSnake(i)->setLastPoint(pos);
  }
  else{
    pos= getSnake(i)->getLastPoint();
    getSnake(i)->getBase().setPosition(pos);
  }

#endif
  getSnake(i)->updateSnake(img,10,eyecup);
}

void SnakeNest::printOrientations(ostream& out,float length){
  for(int i=0;i<nb;i++){
    if(getSnake(i)->isValid()){
      float ori = getSnake(i)->getOrientation(length);
      if (fabs(ori)>90){
	ori -=sign(ori)*180;
      }
      out<<ori<<" ";
    }     
#ifdef WITH_NANS
    else{ 
      out<<"nan ";
    }
#endif
  }
  out<<endl;
}


void SnakeNest::printSegments(ostream& out){
  for(int i=0;i<nb;i++){
    Snake *s = getSnake(i);
    if(s->isValid()){
      Vec2 bot = s->getPoint(s->getSize()-1);
      Vec2 top = s->getPoint(0);
      Vec2 seg = top-bot;
      out<<seg.x<<" "<<seg.y<<" ";
    }
#ifdef WITH_NANS     
    else{ 
      out<<"nan nan ";
    }
#endif
  }
  out<<endl;
} 

void SnakeNest::printLengths(ostream& out){
  for(int i=0;i<nb;i++){
    if(getSnake(i)->isValid()){
      out<<getSnake(i)->getLength()<<" ";
    }  
#ifdef WITH_NANS  
    else{ 
      out<<"nan ";
    }
#endif
  }
  out<<endl;
}
 

int SnakeNest::printSnakes(const char *basefname,int i)const{
  char ofname[100];
  sprintf(ofname,"%s%03d.txt",basefname,i);
  ofstream out(ofname);
  if(!out.is_open()){cerr<<"can't open file "<<ofname<<endl;return 0;}
  printSnakes(out);
  return 1;
}

void SnakeNest::printSnakes(ostream& out)const{
  out<<nb<<endl;
  for(int i=0;i<nb;i++){
    const Snake *s = getSnake(i);  
    out<<(*s)<<endl;
  }
}

void SnakeNest::zoomThresholding(float z){
  if(z>0){
    thresh_w *= 0.8;
    thresh_h *= 0.8;
  }
  if(z<0){
    thresh_w /= 0.8;
    thresh_h /= 0.8;
  }
  cvResizeWindow("thresholding",thresh_w,thresh_h);
  cvShowImage("thresholding",SnakeNest::instance->getImage());
}

void SnakeNest::unzoom(int per)
{
  if(per){
    zoom_factor=1;
    zooming = 0;
    zoomRect = cvRect(0,0,disp->width,disp->height);
  }
  resetZoom();
  // cvShowImage(wname,disp);
  // closeDisplay();
  // initDisplay("HypoPhen",CV_WINDOW_NORMAL,2);
}

void SnakeNest::resetZoom(){
  //  zoom_factor=1;
  // zoomRect = cvRect(0,0,disp->width, disp->height);
  cvResetImageROI(disp);
  cvResetImageROI(idisp);
  cvResetImageROI(img);
  cvResetImageROI(old_img);
  cvResetImageROI(disp_bak);//??
#ifdef GLOBAL_PREPROCESSING
  cvResetImageROI(imgeq);
  cvResetImageROI(old_imgeq);
#endif
}

void SnakeNest::zoom(Vec2 pos,float f){
  zoom_factor *= f;
  if(zoom_factor>=1){
    resetZoom();
    zooming = 0;
  }
  else{
    int lx = disp->width*zoom_factor;
    int ly = disp->height*zoom_factor;
    int nx= max(0,pos.x - lx/2);
    int  ny=max(0,pos.y -ly/2);
    zoomRect = cvRect(nx,ny,lx,ly);
    zoom();
    zooming=1;
  }
}

// deprecated
void SnakeNest::zoom(int col,int row,int frame){
  if(row<nrow && col<ncol){
    int nx = max(0,getSnake(col,row)->minX()-frame);
    int ny = max(0,getSnake(col,row)->minY()-frame);
    int lx = min(disp->width-nx,getSnake(col,row)->maxX()-nx+frame);
    int ly = min(disp->height-ny,getSnake(col,row)->maxY()-ny+frame);
    
    if(ly/(float)lx> disp->height/(float)disp->width){ //keep proportions
      lx = (ly*disp->width)/disp->height;
      if(nx+lx>disp->width){
	nx = disp->width-lx;
      }
    }
    else{
      if(ny+ly>disp->height){
	ny = disp->height-ly;
      }
      ly = (lx*disp->height)/disp->width;
    }
    zoomRect = cvRect(nx,ny,lx,ly);
//     cvSetImageROI(disp,zoomRect);
//     cvSetImageROI(idisp,zoomRect);//for later use
    zoom();
    cvShowImage(wname,disp);
  }
}
void SnakeNest::resizeWindow(float fac){
  win_factor *= fac;
  cvResizeWindow(wname,win_factor*disp->width,win_factor*disp->height);
  //cvShowImage(wname,disp);
  display();
}

void SnakeNest::zoom(){
  cvSetImageROI(disp,zoomRect);
  cvSetImageROI(idisp,zoomRect);//for later use
  cvSetImageROI(img,zoomRect);//for later use
  cvSetImageROI(old_img,zoomRect);//for later use
  cvSetImageROI(disp_bak,zoomRect);
#ifdef GLOBAL_PREPROCESSING
  cvSetImageROI(imgeq,zoomRect);//for later use
  cvSetImageROI(old_imgeq,zoomRect);//for later use
#endif
}

void SnakeNest::discardSnakes(const char *disfname){
  if(disfname){
    ifstream disfile(disfname);
    if(disfile.is_open()){
      int i;
      while(!disfile.eof()){
	disfile>>i;
	if(disfile.good()&&i<=nb){
	  getSnake(i-1)->empty();//no need to do it twice for the last one
	}
      }
      disfile.close();
    }
    else{cout<<"cannot open file "<<disfname<<endl;}
  }
}


void SnakeNest::initCotyledons(const char *cotfname){
  if(cotfname){
      ifstream cotfile(cotfname);
      if(cotfile.is_open()){
	cotfile.close();
	readCotyledons(cotfname);
      }
      else{
	setState(ZOOM);
	for(int i=0;i<nb;i++){
	  if(getSnake(i)->isValid()){
	    getSnake(i)->draw(disp,CV_RGB(0,0,255));
	    selected_snake=i;
	    zoom(i%ncol,i/ncol,20);
	    //getSnake(i)->draw(disp,CV_RGB(255,0,0));
	    cvWaitKey();
	    resetZoom();
	  }
	}
	writeCotyledons(cotfname);
	setState(CORRECT);
      }
  }
}

void SnakeNest::setState(NestState s){
  if(s==state)return;
  switch(s){
  case ZOOM:
    closeDisplay();
    initDisplay("HypoPhen",0,0);
    break;
  case CORRECT:
    if(state==ZOOM){
      unzoom();
    }    
    break;
  default:;
  }
  state=s;
}


int SnakeNest::setImageFileName(char *outimgfile){ 
  strcpy(baseimgfile,strtok(outimgfile,"."));
  char *p = strtok(NULL,".");
  if(p){
    strcpy(imgfileformat,p);
    saveImages=1;
    return 1;
  }
  else{
    cerr<<"cannot find extension in filename "<<outimgfile<<endl;
    return 0;
  }
}

void SnakeNest::saveImage(int n){
  char fname[80];
  sprintf(fname,"%s%03d.%s",baseimgfile,n,imgfileformat);
  cvSaveImage(fname,disp);
}


int SnakeNest::openResFiles(const char *pofname, ofstream& or_out,
			    ofstream& len_out,ofstream& seg_out,
			    ofstream& tip_out){
 char orfname[100];
  char lenfname[100];
  char segfname[100];
  char tipfname[100];
  if(!keep_trace){
    sprintf(orfname,"%s%s",pofname,"_or.txt");
    sprintf(lenfname,"%s%s",pofname,"_len.txt");
    sprintf(segfname,"%s%s",pofname,"_seg.txt");
    sprintf(tipfname,"%s%s",pofname,"_tip.txt");
    or_out.open(orfname);
    if(!or_out.is_open()){cout<<"cannot open file "<<orfname<<endl;return 0;} 
    len_out.open(lenfname);
    if(!len_out.is_open()){cout<<"cannot open file "<<lenfname<<endl;return 0;} 
    seg_out.open(segfname);
    if(!seg_out.is_open()){cout<<"cannot open file "<<segfname<<endl;return 0;}
    tip_out.open(tipfname);
    if(!tip_out.is_open()){cout<<"cannot open file "<<tipfname<<endl;return 0;}
  }      
  return 1;
}
int SnakeNest::localThresholding(int i){
  if(i>=0){//local threshold
    selected_snake =i;
    zoom(i%ncol,i/ncol,10);
    bw_threshold = bw_thresholds[i];
  } 
   threshold();
   
   //   cvNamedWindow("thresholding",CV_WINDOW_AUTOSIZE);
   cvNamedWindow("thresholding",0);
   cvResizeWindow("thresholding",thresh_w,thresh_h);
   cvShowImage("thresholding",img); 
   cvCreateTrackbar("threshold","thresholding",&bw_threshold,255,threshold_tbar);
   cvSetMouseCallback("thresholding",on_mouse_thresholding);
   
  return 0;
}

void SnakeNest::getThresholdValues(int j){
  char c=0;
  if(j>=0){
    if(getSnake(j)->isValid()){
      localThresholding(j);
      while('g'!=cvWaitKey(5)){};     
      unzoom();
      getSnake(j)->updateSnake(img,10,EYECUP);
    }
  }
  else{
    switch(j){
    case -1: //all hypocotyls
      for(int i=0;i<nb;i++){
	if(getSnake(i)->isValid()){
	  localThresholding(i);
	  while(' '!= cvWaitKey(5)){};     
	  unzoom();
	}
      }
      break;
    case -2: //global threshold
      global_threshold=0;
      localThresholding(-1);
      while(c==0){
	c=cvWaitKey(5);
	switch(c){
	case 'd':
	  cvDestroyWindow("thresholding");
	  global_threshold=-1;
	  getThresholdValues(-1);
	  break;
	case 'o': zoomThresholding(1);c=0;break;
	case 'i': zoomThresholding(-1);c=0;break;
	case ' ':break;
	default:c=0;break;
	}
      }
  default:;
    }
  }
  cvDestroyWindow("thresholding");
}

#ifdef WITH_HISTORY
void SnakeNest::saveSnakes(uint tim){
   for(int i=0;i<nb;i++){
      if(getSnake(i)->isValid()){
	history.setSnake(getSnake(i),i,tim);
      }
   }
}



void SnakeNest::backToTime(uint tim){
  for(int i=0;i<nb;i++){
    Snake *s = getSnake(i);    
    if(history.getSnakeSize(i,tim)){
      s->validate();
      history.getSnake(i,tim,*s);
      s->getCotyledon().setPosition(s->getPoint(0));
#ifdef WITH_BASE
      s->getBase().setPosition(s->getLastPoint());
#endif
    }
    else{
      s->invalidate();
    }
  }
}

#endif

void SnakeNest::run(int neb_size,const char *imfname,const char *pofname){
  ofstream or_out,len_out,seg_out,tip_out;  
  if(!openResFiles(pofname,or_out,len_out,seg_out,tip_out)){return;} 
  setNeighborhood(neb_size);
  initGrabber(imfname);
  cout<<"initializing display"<<endl;
  //  initDisplay("HypoPhen",CV_WINDOW_AUTOSIZE,2);//for a detection trackbar
  initDisplay("HypoPhen",CV_WINDOW_NORMAL,2);//for a detection trackbar
  nextFrame();
  basicInitSnakes();
  invalidateAll(); 
  display(CV_RGB(255,0,0));
  setState(CORRECT);
  int stop=0;
  cout<<"press 'i' to enlarge the window and 'o' to make it smaller"<<endl;
  cout<<"select the hypocotyls with (ctl)alt-right click"<<endl;
  cout<<"press <space> when you are done"<<endl;
  while(!stop){
    char c =cvWaitKey(5);
    stop= c==' ';
    if(c=='o'){
      resizeWindow(0.8);
    }
    if(c=='i'){
      resizeWindow(1/0.8);
    }
  }
  setState(CALIB);
  unzoom(1);
  cout<<"calibrate the thresholding using the slider"<<endl;
  cout<<"press 'i' to enlarge the window and 'o' to make it smaller"<<endl;
  cout<<"press 'd' to calibrate each hypocotyl separately"<<endl;
  cout<<"press <space> when you are done"<<endl;
  getThresholdValues(-2);
  for(int i=0;i<nb;i++){
    // selected_snake=i;	
    if(getSnake(i)->isValid()){	
      bwUpdateSnake(i,10,EYECUP);
    }
  }
  display(CV_RGB(255,0,0));
  setState(GET_GROUND);
  cout<<"use the mouse to set the bottom of hypocotyl rows or press <space> to skip (skipping recommended)"<<endl;
  while(' '!=cvWaitKey(5)){}
  cout<<"********"<<endl;
  cout<<"you can now monitor the semi-automatic hypocotyl phenotyping"<<endl;
  cout<<"press <space> to go to the next image"<<endl;
  cout<<"refer to the manual for details"<<endl;
  cout<<"********"<<endl;
  setState(CORRECT);
  stop=0;
  do{  
    for(int i=0;i<nb;i++){
      if(getSnake(i)->isValid()){
	//	selected_snake=i;	
	bwUpdateSnake(i,10,EYECUP);
	//	cout<<"cot3 "<<i<<" "<<getSnake(i)->getCotyledon().getPosition()<<endl;
	//	getSnake(i)->drawCotyledon(disp);
      }
    }
    display(CV_RGB(255,0,0));
    int nxt= 0;
    while(!nxt){
     char c =cvWaitKey(10);
       switch(c){
#ifdef WITH_HISTORY
       case 'b': 
	 //	 cout<<"going back"<<endl;
	 backToTime(grabber->GetCurrentTimeIndex()-2);
	 //	 cout<<"one image back"<<endl;
	 grabber->SetFrame(grabber->GetCurrentFrame()-2);
	 nextFrame();
	 //	 cout<<"displaying"<<endl;
	 display();
	 break;
       case 'n':
	 if(presentTime()){
	   cout<<"back to present"<<endl;
	   nxt=1;
	 }
	 else{
	   backToTime(grabber->GetCurrentTimeIndex());
	   nextFrame();
	   display();
	 }
	 break;
       case 'v'://validate
	 if(!presentTime()){
	   uint tim = grabber->GetCurrentTimeIndex()-1;
	   saveSnakes(tim);
	   if(saveImages){saveImage(tim);}
	   cout<<"saving analysis of image "<<grabber->GetCurrentFrame()-1<<endl;
	 }
	 else{
	   nxt=1;
	 }
	 break;
#endif
       case 'l':
	 if(oriwin){
	   closeDisplay();
	   //	   initDisplay(wname,CV_WINDOW_AUTOSIZE,2);
	   initDisplay(wname,CV_WINDOW_NORMAL,2);
	   display();
	 }	 
	 else{
	   cvCreateTrackbar("orientation parameter",wname,
			    &orient_length,max_orient_length,orient_tbar);
	   orient_tbar(orient_length);
	 }
	 oriwin = !oriwin;
	 break;
       case 'o': //smaller window
	 resizeWindow(0.8);
	 break;
       case 'i': //bigger window
	 resizeWindow(1/0.8);
	 break;
       case 'z'://unzoom
	 unzoom(1);
	 display();
	 break;
       case 'q':stop=1;//quit programm
       case ' ':nxt=1; //continue
       default:;
       }
     }// waits that the users hits the SPACE key
     if(stop){break;}
     uint tim = grabber->GetCurrentTimeIndex()-1;
#ifndef WITH_HISTORY
     if(keep_trace){
       printSnakes(pofname,grabber->GetCurrentTimeIndex()-1);
     }
     else{
       printLengths(len_out);
       printOrientations(or_out,orient_length);
       printSegments(seg_out);
     }     
#else
     saveSnakes(tim);
#endif
     if(saveImages){saveImage(tim);}
  }while(nextFrame());

#ifdef WITH_HISTORY
  history.printAllLengths(len_out);
  history.printAllOrientations(or_out,orient_length);
  history.printAllSegments(seg_out);
  history.printAllTips(tip_out);
#endif
  len_out.close();
  or_out.close();
  seg_out.close();
  tip_out.close();
  grabber->Kill();
}





#ifdef TUNE_COTYLEDON_TRACKING

void SnakeNest::initCotyledonDebugDisplay(const char *win_name){
  disp_bak = cvCloneImage(disp); 
  idisp = cvCloneImage(disp);
  strcpy(wname,win_name);
  cvNamedWindow(wname,1);
  cvMoveWindow(wname,100,300);
  
  cvSetMouseCallback(wname,on_mouse,this);
  cvCreateTrackbar("smoothness",win_name,&modulation,max_modulation,modulation_tbar);
  
  cvNamedWindow("cot1",0);
  cvMoveWindow("cot1",100,000);
  
  cvNamedWindow("cot2",0);
  cvMoveWindow("cot2",600,000);
}

void SnakeNest::testCotyledonTracking(char *imfname, char *cotfname, int i){
  initGrabber(imfname);
  initCotyledonDebugDisplay("hypotrack");
  nextFrame();
  display(CV_RGB(255,0,0));
  initSnakes();
  //  discardSnakes(disfile);
  updateSnakes();
  display(CV_RGB(255,0,0));
  while(' '!=cvWaitKey()){}
  setState(CORRECT);
  initCotyledons(cotfname);
  int stop=0;
  int first =1;
  Vec2 trans;
  do{  
    selected_snake=i;	
    if(getSnake(i)->isValid()){
      //if(!first) getSnake(i)->updateCotyledon(img,3);
 
      if(!first){ trans = getSnake(i)->updateCotyledonOF(img,old_img);}
      getSnake(i)->drawCotyledon(disp);
   
 getSnake(i)->getCotyledon().hide(img);
      getSnake(i)->translatePoint(0,trans.round());

#ifdef SNAKE_EYECUP
	getSnake(i)->updateSnake(img,10,EYECUP);//was 20
#else
	getSnake(i)->updateSnake(img,10);//was 20
#endif   
    }
    first=0;
    display(CV_RGB(255,0,0));
    int nxt= 0;
    while(!nxt){
       char c =cvWaitKey(10);
       switch(c){
       case 't':translateAll(Vec2(0,10));updateSnakes();display();break;
       case 'q':stop=1;//quit programm
       case ' ':nxt=1; //continue
       default:;
       }
     }// waits that the users hits the SPACE key
     if(stop){break;}
  }while(nextFrame());
  grabber->Kill();
}


#endif


#define SNAKE_NEST_MAIN
#ifdef SNAKE_NEST_MAIN

int usage(){
  cout<<"usage: ./hypoPhen <image_file_name> <nb_rows> <nb_cols> <out_file_name> [imagefile=<outimagefile>] [--trace]"
      <<endl; 
  return 0;
}


int main(int argc, char *argv[]){
  char imgfilename[200];
  char resfilename[200];
  int rows;
  int cols;
  if(argc==1){
    rows=1;
    cols = 6;
    strcpy(imgfilename,argv[0]);
    strcpy(imgfilename+strlen(imgfilename)-EXEC_LENGTH,"./images/emilie/emilie00.png");
    strcpy(resfilename,argv[0]);
    strcpy(resfilename+strlen(resfilename)-EXEC_LENGTH,"./results");
   
  }
  else{
    if(argc <5){
      return usage();
    }
    strcpy(imgfilename,argv[1]);
    strcpy(resfilename,argv[4]);
    rows = atoi(argv[2]);
    cols = atoi(argv[3]);
  }
  int npts = 10;
  char *cotfile=NULL;
  char *outimgfile=NULL;
  SnakeNest *pnest = new SnakeNest(cols,rows,npts);
  SnakeNest& nest = *pnest;
  for(int i=5;i<argc;i++){
    char *p; 
    p=strtok(argv[i],"=");
    if(!strcmp(p,"cot")){
      cotfile = strtok(NULL,"=");
      cout<<cotfile<<endl;
    }
    else if(!strcmp(p,"imagefile")){
      outimgfile = strtok(NULL,"=");
      cout<<outimgfile<<endl;
      if(!nest.setImageFileName(outimgfile)){return 0;}
    }
    else if(!strcmp(p,"--trace")){
      cout<<"tracing"<<endl;
      nest.keepTrace();
    }
    else if(!strcmp(p,"--growth")){
      int disp = atoi(strtok(NULL,"="));
      cout<<"max growth: "<<disp<<endl;
      OpenCotyledon::setMaxDisplacement(disp);
    }

  }
  nest.run(5,imgfilename,resfilename);
  delete pnest;
  return 0;
}
#endif
